@@ -8,11 +8,15 @@
 #'
 #' @examples
 #' #none
-convert_color = function(color) {
-  if(class(color) == "character") {
+convert_color = function(color, as_hex = FALSE) {
+  if(inherits(color,"character")) {
     color = as.vector(col2rgb(color))/255
   } 
   assertthat::assert_that(all(color <= 1))
   assertthat::assert_that(all(color >= 0))
-  color
+  if(as_hex) {
+    paste0("#",paste0(format(as.hexmode(round(color*255,0)),width=2),collapse=""),collapse="")
+  } else {
+    color
+  }
 }
\ No newline at end of file
