@@ -7,12 +7,20 @@
 #' @param height Default `400`. Height of the render, in pixels.
 #' @param fov Default `20`. Field of view, in degrees. If this is zero, the camera will use an orthographic projection. The size of the plane
 #' used to create the orthographic projection is given in argument `ortho_dimensions`.
-#' @param samples Default `100`. The maximum number of samples for each pixel.
+#' @param samples Default `100`. The maximum number of samples for each pixel. If this is a length-2
+#' vector and the `sample_method` is `stratified`, this will control the number of strata in each dimension.
+#' The total number of samples in this case will be the product of the two numbers.
 #' @param min_variance Default `0.00005`. Minimum acceptable variance for a block of pixels for the 
 #' adaptive sampler. Smaller numbers give higher quality images, at the expense of longer rendering times.
 #' If this is set to zero, the adaptive sampler will be turned off and the renderer
 #' will use the maximum number of samples everywhere.
 #' @param min_adaptive_size Default `8`. Width of the minimum block size in the adaptive sampler.
+#' @param sample_method Default `random`. The type of sampling method used to generate
+#' random numbers. The other option is `stratified`, which can improve the render quality (at the cost
+#' of increased time allocating the random samples).
+#' @param max_depth Default `50`. Maximum number of bounces a ray can make in a scene.
+#' @param roulette_active_depth Default `10`. Number of ray bounces until a ray can stop bouncing via
+#' Russian roulette.
 #' @param ambient_light Default `FALSE`, unless there are no emitting objects in the scene. 
 #' If `TRUE`, the background will be a gradient varying from `backgroundhigh` directly up (+y) to 
 #' `backgroundlow` directly down (-y).
@@ -51,6 +59,8 @@
 #' @param environment_light Default `NULL`. An image to be used for the background for rays that escape
 #' the scene. Supports both HDR (`.hdr`) and low-dynamic range (`.png`, `.jpg`) images.
 #' @param rotate_env Default `0`. The number of degrees to rotate the environment map around the scene.
+#' @param intensity_env Default `1`. The amount to increase the intensity of the environment lighting. Useful
+#' if using a LDR (JPEG or PNG) image as an environment map.
 #' @param debug_channel Default `none`. If `depth`, function will return a depth map of rays into the scene 
 #' instead of an image. If `normals`, function will return an image of scene normals, mapped from 0 to 1.
 #' If `uv`, function will return an image of the uv coords. If `variance`, function will return an image 
@@ -66,34 +76,36 @@
 #'
 #' @examples
 #' #Generate a large checkered sphere as the ground
-#' scene = generate_ground(depth=-0.5, material = diffuse(color="white", checkercolor="darkgreen"))
 #' \donttest{
+#' scene = generate_ground(depth=-0.5, material = diffuse(color="white", checkercolor="darkgreen"))
 #' render_scene(scene,parallel=TRUE,samples=500)
 #' }
 #' 
 #' #Add a sphere to the center
+#' \donttest{
 #' scene = scene %>%
 #'   add_object(sphere(x=0,y=0,z=0,radius=0.5,material = diffuse(color=c(1,0,1))))
-#' \donttest{
 #' render_scene(scene,fov=20,parallel=TRUE,samples=500)
 #' }
 #' 
 #' #Add a marbled cube 
+#' \donttest{
 #' scene = scene %>%
 #'   add_object(cube(x=1.1,y=0,z=0,material = diffuse(noise=3)))
-#' \donttest{
 #' render_scene(scene,fov=20,parallel=TRUE,samples=500)
 #' }
 #' 
-#' #Add a metallic gold sphere
+#' #Add a metallic gold sphere, using stratified sampling for a higher quality render
+#' \donttest{
 #' scene = scene %>%
 #'   add_object(sphere(x=-1.1,y=0,z=0,radius=0.5,material = metal(color="gold",fuzz=0.1)))
-#' \donttest{
-#' render_scene(scene,fov=20,parallel=TRUE,samples=500)
+#' render_scene(scene,fov=20,parallel=TRUE,samples=500, sample_method = "stratified")
 #' }
 #' 
 #' #Lower the number of samples to render more quickly (here, we also use only one core).
+#' \donttest{
 #' render_scene(scene, samples=4)
+#' }
 #' 
 #' #Add a floating R plot using the iris dataset as a png onto a floating 2D rectangle
 #' 
@@ -141,19 +153,22 @@
 #'on.exit(par(old.par))
 #'par(mfrow=c(5,6))
 #'for(i in 1:30) {
-#'  render_scene(scene, samples=5,
+#'  render_scene(scene, samples=16, 
 #'    lookfrom = c(xpos[i],1.5,zpos[i]),lookat = c(0,0.5,0), parallel=TRUE)
 #'}
 #'}
 render_scene = function(scene, width = 400, height = 400, fov = 20, 
                         samples = 100, min_variance = 0.00005, min_adaptive_size = 8,
-                        ambient_light = FALSE,
+                        sample_method = "random",
+                        max_depth = 50, roulette_active_depth = 10,
+                        ambient_light = FALSE, 
                         lookfrom = c(0,1,10), lookat = c(0,0,0), camera_up = c(0,1,0), 
                         aperture = 0.1, clamp_value = Inf,
                         filename = NULL, backgroundhigh = "#80b4ff",backgroundlow = "#ffffff",
                         shutteropen = 0.0, shutterclose = 1.0, focal_distance=NULL, ortho_dimensions = c(1,1),
                         tonemap ="gamma", bloom = TRUE, parallel=TRUE, 
-                        environment_light = NULL, rotate_env = 0, debug_channel = "none",
+                        environment_light = NULL, rotate_env = 0, intensity_env = 1,
+                        debug_channel = "none",
                         progress = interactive(), verbose = FALSE) { 
   if(verbose) {
     currenttime = proc.time()
@@ -204,9 +219,9 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
   shapevec = unlist(lapply(tolower(scene$shape),switch,
                           "sphere" = 1,"xy_rect" = 2, "xz_rect" = 3,"yz_rect" = 4,"box" = 5, "triangle" = 6, 
                           "obj" = 7, "objcolor" = 8, "disk" = 9, "cylinder" = 10, "ellipsoid" = 11,
-                          "objvertexcolor" = 12))
+                          "objvertexcolor" = 12, "cone" = 13))
   typevec = unlist(lapply(tolower(scene$type),switch,
-                          "diffuse" = 1,"metal" = 2,"dielectric" = 3, "oren-nayar" = 4, "light" = 5))
+                          "diffuse" = 1,"metal" = 2,"dielectric" = 3, "oren-nayar" = 4, "light" = 5, "microfacet" = 6, "glossy" = 7))
   sigmavec = unlist(scene$sigma)
   
   assertthat::assert_that(tonemap %in% c("gamma","reinhold","uncharted", "hbd", "raw"))
@@ -218,12 +233,18 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
   checkeredlist = scene$checkercolor
   checkeredbool = purrr::map_lgl(checkeredlist,.f = ~all(!is.na(.x)))
   
+  #glossy 
+  glossyinfo = scene$glossyinfo
+
   #gradient handler
   gradient_info = list()
   gradient_info$gradient_colors = scene$gradient_color
   gradient_info$isgradient = purrr::map_lgl(gradient_info$gradient_colors,.f = ~all(!is.na(.x)))
   gradient_info$gradient_trans = scene$gradient_transpose
-  
+  gradient_info$is_world_gradient = scene$world_gradient
+  gradient_info$gradient_control_points = scene$gradient_point_info
+  gradient_info$type = unlist(lapply(tolower(scene$gradient_type),switch,
+                                     "hsv" = TRUE, "rgb" = FALSE, FALSE))
   #noise handler
   noisebool = purrr::map_lgl(scene$noise, .f = ~.x > 0)
   noisevec = scene$noise
@@ -270,6 +291,7 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
     }
   }
   image_tex_bool = image_tex_bool | image_filename_bool
+  image_repeat = scene$image_repeat
   #alpha texture handler
   alpha_array_list = scene$alphaimage
   alpha_tex_bool = purrr::map_lgl(alpha_array_list,.f = ~is.array(.x[[1]]))
@@ -308,6 +330,44 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
   alphalist = list()
   alphalist$alpha_temp_file_names = alpha_temp_file_names
   alphalist$alpha_tex_bool = alpha_tex_bool
+  
+  #bump texture handler
+  bump_array_list = scene$bump_texture
+  bump_tex_bool = purrr::map_lgl(bump_array_list,.f = ~is.array(.x[[1]]))
+  bump_filename_bool = purrr::map_lgl(bump_array_list,.f = ~is.character(.x[[1]]))
+  bump_temp_file_names = purrr::map_chr(bump_tex_bool,.f = ~ifelse(.x, tempfile(fileext = ".png"),""))
+  for(i in 1:length(bump_array_list)) {
+    if(bump_tex_bool[i]) {
+      bump_dims = dim(bump_array_list[[i]][[1]])
+      if(length(bump_dims) == 2) {
+        temp_array = array(0, dim = c(bump_dims,3))
+        temp_array[,,1] = bump_array_list[[i]][[1]]
+        temp_array[,,2] = bump_array_list[[i]][[1]]
+        temp_array[,,3] = bump_array_list[[i]][[1]]
+        bump_dims = c(bump_dims,3)
+      } else {
+        temp_array = bump_array_list[[i]][[1]]
+      }
+      if(bump_dims[3] == 4) {
+        png::writePNG(fliplr(aperm(temp_array[,,1:3],c(2,1,3))),bump_temp_file_names[i])
+      } else if(bump_dims[3] == 3){
+        png::writePNG(fliplr(aperm(temp_array,c(2,1,3))),bump_temp_file_names[i])
+      }
+    }
+    if(bump_filename_bool[i]) {
+      if(any(!file.exists(path.expand(bump_array_list[[i]][[1]])) & nchar(bump_array_list[[i]][[1]]) > 0)) {
+        stop(paste0("Cannot find the following texture file:\n",
+                    paste(bump_array_list[[i]][[1]], collapse="\n")))
+      }
+      bump_temp_file_names[i] = path.expand(bump_array_list[[i]][[1]])
+    }
+  }
+  bump_tex_bool = bump_tex_bool | bump_filename_bool
+  bump_intensity = scene$bump_intensity
+  alphalist$bump_temp_file_names = bump_temp_file_names
+  alphalist$bump_tex_bool = bump_tex_bool
+  alphalist$bump_intensity = bump_intensity
+  
   #movement handler
   if(shutteropen == shutterclose) {
     movingvec = rep(FALSE,length(movingvec))
@@ -365,17 +425,6 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
   
   assertthat::assert_that(all(c(length(position_list$xvec),length(position_list$yvec),length(position_list$zvec),length(rvec),length(typevec),length(proplist)) == length(position_list$xvec)))
   assertthat::assert_that(all(!is.null(typevec)))
-  for(i in 1:length(position_list$xvec)) {
-    if(typevec[i] == 1) {
-      if(shapevec[i] == 1) {
-        # assertthat::assert_that(length(proplist[[i]]) == 3)
-      }
-    } else if (typevec[i] == 2) {
-      # assertthat::assert_that(length(proplist[[i]]) == 4)
-    } else if (typevec[i] == 3) {
-      # assertthat::assert_that(length(proplist[[i]]) == 7)
-    } 
-  }
   assertthat::assert_that(length(lookfrom) == 3)
   assertthat::assert_that(length(lookat) == 3)
   if(is.null(focal_distance)) {
@@ -404,7 +453,17 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
     buildingtime = proc.time() - currenttime
     cat(sprintf("%0.3f seconds \n",buildingtime[3]))
   }
+  sample_method = unlist(lapply(tolower(sample_method),switch,
+                                "random" = 0,"stratified" = 1, 0))
+  
   camera_info = list()
+  strat_dim = c()
+  if(length(samples) == 2) {
+    strat_dim = samples
+    samples = samples[1]*samples[2]
+  } else {
+    strat_dim = rep(min(floor(sqrt(samples)),8),2)
+  }
   
   camera_info$nx = width
   camera_info$ny = height
@@ -418,11 +477,20 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
   camera_info$shutterclose = shutterclose
   camera_info$ortho_dimensions = ortho_dimensions
   camera_info$focal_distance = focal_distance
-  camera_info$toneval = toneval
+  camera_info$max_depth = max_depth
+  camera_info$roulette_active_depth = roulette_active_depth
+  camera_info$sample_method = sample_method
+  camera_info$stratified_dim = strat_dim
   
-  #Material ID handler
+  assertthat::assert_that(max_depth > 0)
+  assertthat::assert_that(roulette_active_depth > 0)
+  
+  
+  #Material ID handler; these must show up in increasing order.  Note, this will
+  #cause problems if `match` is every changed to return doubles when matching in
+  #long vectors as has happened with `which` recently.
   material_id = scene$material_id
-  material_id = as.integer(as.factor(material_id))-1
+  material_id = as.integer(match(material_id, unique(material_id)) - 1L)
   material_id_bool = !is.na(scene$material_id)
   
   if(min_adaptive_size < 1) {
@@ -455,10 +523,11 @@ render_scene = function(scene, width = 400, height = 400, fov = 20,
                              fileinfo = objfilenamevec, filebasedir = objbasedirvec,
                              progress_bar = progress, numbercores = numbercores, 
                              hasbackground = hasbackground, background = backgroundstring, scale_list = scale_factor,
-                             sigmavec = sigmavec, rotate_env = rotate_env,
+                             sigmavec = sigmavec, rotate_env = rotate_env, intensity_env = intensity_env,
                              verbose = verbose, debug_channel = debug_channel,
                              shared_id_mat=material_id, is_shared_mat=material_id_bool, 
-                             min_variance = min_variance, min_adaptive_size = min_adaptive_size) 
+                             min_variance = min_variance, min_adaptive_size = min_adaptive_size, 
+                             glossyinfo = glossyinfo, image_repeat = image_repeat) 
   full_array = array(0,c(ncol(rgb_mat$r),nrow(rgb_mat$r),3))
   full_array[,,1] = flipud(t(rgb_mat$r))
   full_array[,,2] = flipud(t(rgb_mat$g))
