@@ -1,12 +1,17 @@
 #define STB_IMAGE_IMPLEMENTATION 
 
+
+#ifndef FLOATDEF
+#define FLOATDEF
 #ifdef RAY_FLOAT_AS_DOUBLE
 typedef double Float;
 #else
 typedef float Float;
 #endif 
+#endif
 
 #include "vec3.h"
+#include "vec2.h"
 #include "mathinline.h"
 #include "camera.h"
 #include "float.h"
@@ -16,11 +21,19 @@ typedef float Float;
 #include "tonemap.h"
 #include "infinite_area_light.h"
 #include "adaptivesampler.h"
+#include "sampler.h"
 using namespace Rcpp;
 // [[Rcpp::plugins(cpp11)]]
 // [[Rcpp::depends(RcppThread)]]
 #include "RcppThread.h"
 
+// #define DEBUG
+
+#ifdef DEBUG
+#include <iostream>
+#include <fstream>
+#endif
+
 using namespace std;
 
 inline vec3 de_nan(const vec3& c) {
@@ -31,141 +44,86 @@ inline vec3 de_nan(const vec3& c) {
   return(temp);
 }
 
-vec3 color(const ray& r, hitable *world, hitable *hlist, int depth, bool tonemap, random_gen& rng) {
-  hit_record hrec;
-  if(world->hit(r, 0.001, FLT_MAX, hrec, rng)) { //generated hit record, world space
-    scatter_record srec;
-    vec3 emitted = hrec.mat_ptr->emitted(r, hrec, hrec.u, hrec.v, hrec.p);
-    float pdf_val;
-    if(depth < 50 && hrec.mat_ptr->scatter(r, hrec, srec, rng)) { //generates scatter record, world space
-      if(srec.is_specular) { //returns specular ray
-        // srec.specular_ray.origin() = offset_ray(srec.specular_ray.origin(), hrec.normal);
-        return(srec.attenuation * 
-               color(srec.specular_ray, world, 
-                     hlist, depth + 1, tonemap, rng));
-      }
-      hitable_pdf p_imp(hlist, hrec.p); //creates pdf of all objects to be sampled
-      mixture_pdf p(&p_imp, srec.pdf_ptr); //creates mixture pdf of surface intersected at hrec.p and all sampled objects/lights
-      
-      //Generates a scatter direction (with origin hrec.p) from the mixture 
-      //and saves surface normal from light to use in pdf_value calculation
-      //(along with the scatter direction)
-      //Translates the world space point into object space point, generates ray assuring intersection, and then translates 
-      //ray back into world space
-      // ray scattered = ray(offset_ray(hrec.p,hrec.normal), p.generate(rng), r.pri_stack, r.time()); //scatters a ray from hit point to direction
-      ray scattered = ray(hrec.p, p.generate(rng), r.pri_stack, r.time()); //scatters a ray from hit point to direction
-      pdf_val = p.value(scattered.direction(), rng); //generates a pdf value based the intersection point and the mixture pdf
-      return(emitted + srec.attenuation *
-             hrec.mat_ptr->scattering_pdf(r, hrec, scattered) *
-             color(scattered, world,
-                  hlist, depth + 1, tonemap, rng) / pdf_val);
-    } else {
-      return(emitted);
-    }
-  } else {
-    return(vec3(0,0,0));
-  }
-}
 
-vec3 color_amb(const ray& r, hitable *world, hitable *hlist, int depth,
-           const vec3& backgroundhigh, const vec3& backgroundlow, bool tonemap, random_gen& rng) {
-  hit_record hrec;
-  if(world->hit(r, 0.001, FLT_MAX, hrec, rng)) {
-    scatter_record srec;
-    vec3 emitted = hrec.mat_ptr->emitted(r, hrec, hrec.u, hrec.v,hrec.p);
-    float pdf_val;
-    if(depth < 50 && hrec.mat_ptr->scatter(r, hrec, srec, rng)) {
-      if(srec.is_specular) {
-        return(srec.attenuation * 
-               color_amb(srec.specular_ray, world, hlist, depth + 1, 
-                         backgroundhigh,backgroundlow,tonemap, rng));
+vec3 color(const ray& r, hitable *world, hitable_list *hlist,
+           size_t max_depth, size_t roulette_activate, random_gen& rng, Sampler* sampler) {
+#ifdef DEBUG
+  ofstream myfile;
+  myfile.open("rays.txt", ios::app | ios::out);
+#endif
+  vec3 final_color(0,0,0);
+  vec3 throughput(1,1,1);
+  float prev_t = 1;
+  ray r1 = r;
+  ray r2 = r;
+  bool diffuse_bounce = false;
+  for(size_t i = 0; i < max_depth; i++) {
+    hit_record hrec;
+    if(world->hit(r2, 0.001, FLT_MAX, hrec, rng)) { //generated hit record, world space
+#ifdef DEBUG
+      myfile << i << ", " << r2.A << " ";
+      myfile << ", " << hrec.p << "\n ";
+#endif
+      scatter_record srec;
+      final_color += throughput * hrec.mat_ptr->emitted(r2, hrec, hrec.u, hrec.v, hrec.p);
+      if(throughput.x() == 0 && throughput.y() == 0 && throughput.z() == 0) {
+        return(vec3(0,0,0));
       }
-      hitable_pdf p_imp(hlist, hrec.p);
-      mixture_pdf p(&p_imp, srec.pdf_ptr);
-      ray scattered = ray(hrec.p, p.generate(rng), r.pri_stack, r.time());
-      pdf_val = p.value(scattered.direction(), rng);
-      return(emitted + srec.attenuation * 
-             hrec.mat_ptr->scattering_pdf(r, hrec, scattered) *  
-             color_amb(scattered, world, hlist, depth + 1, 
-                       backgroundhigh,backgroundlow, tonemap, rng) / pdf_val);
-    } else {
-      vec3 unit_direction = unit_vector(r.direction());
-      float t = 0.5f * (unit_direction.y() + 1.0f);
-      return(emitted + (1.0f - t) * backgroundlow + t * backgroundhigh);
-    }
-  } else {
-    vec3 unit_direction = unit_vector(r.direction());
-    float t = 0.5f * (unit_direction.y() + 1.0);
-    return (1.0 - t) * backgroundlow + t * backgroundhigh;
-  }
-}
-
-vec3 color_uniform(const ray& r, hitable *world, int depth, bool tonemap, random_gen& rng, texture* background_texture) {
-  hit_record hrec;
-  if(world->hit(r, 0.001, FLT_MAX, hrec, rng)) {
-    scatter_record srec;
-    vec3 emitted = hrec.mat_ptr->emitted(r, hrec, hrec.u, hrec.v,hrec.p);
-    float pdf_val;
-    if(depth < 50 && hrec.mat_ptr->scatter(r, hrec, srec, rng)) {
-      if(srec.is_specular) {
-        return(srec.attenuation * color_uniform(srec.specular_ray, world, depth + 1, 
-                                                tonemap, rng, background_texture));
+      if(i > roulette_activate) {
+        float t = std::max(throughput.x(), std::max(throughput.y(), throughput.z()));
+        //From Szecsi, Szirmay-Kalos, and Kelemen
+        float prob_continue = std::min(1.0f, std::sqrt(t/prev_t));
+        prev_t = t;
+        if(rng.unif_rand() > prob_continue) {
+          return(final_color);
+        }
+        throughput *= 1 / prob_continue;
       }
-      cosine_pdf p(hrec.normal);
-      ray scattered = ray(hrec.p, p.generate(rng), r.pri_stack, r.time());
-      pdf_val = p.value(scattered.direction(), rng);
-      return(emitted + srec.attenuation * hrec.mat_ptr->scattering_pdf(r, hrec, scattered) *  
-             color_uniform(scattered, world, depth + 1, tonemap, rng, background_texture) / pdf_val);
-    } else {
-      return(emitted);
-    }
-  } else {
-    vec3 unit_direction = unit_vector(r.direction());
-    float phi = atan2(unit_direction.x(),unit_direction.z());
-    float u = 0.5 + phi / (2*M_PI);
-    float v = 0.5 * (1.0 + unit_direction.y());
-    return(background_texture->value(u, v, unit_direction));
-  }
-}
+      float pdf_val;
+      if(hrec.mat_ptr->scatter(r2, hrec, srec, rng)) { //generates scatter record, world space
+        if(srec.is_specular) { //returns specular ray
+          r2 = srec.specular_ray;
+          throughput *= srec.attenuation;
+          continue;
+        }
+        hitable_pdf p_imp(hlist, hrec.p); //creates pdf of all objects to be sampled
+        mixture_pdf p(&p_imp, srec.pdf_ptr); //creates mixture pdf of surface intersected at hrec.p and all sampled objects/lights
+        
+        //Generates a scatter direction (with origin hrec.p) from the mixture 
+        //and saves surface normal from light to use in pdf_value calculation
+        //(along with the scatter direction)
+        //Translates the world space point into object space point, generates ray assuring intersection, and then translates 
+        //ray back into world space
+        vec3 offset_p = offset_ray(hrec.p-r2.A, hrec.normal) + r2.A;
 
-vec3 color_amb_uniform(const ray& r, hitable *world, int depth,
-               const vec3& backgroundhigh, const vec3& backgroundlow, bool tonemap, random_gen& rng) {
-  hit_record hrec;
-  if(world->hit(r, 0.001, FLT_MAX, hrec, rng)) {
-    scatter_record srec;
-
-    vec3 emitted = hrec.mat_ptr->emitted(r, hrec, hrec.u, hrec.v,hrec.p);
-    float pdf_val;
-    if(depth < 50 && hrec.mat_ptr->scatter(r, hrec, srec, rng)) {
-      if(srec.is_specular) {
-        return(srec.attenuation * 
-               color_amb_uniform(srec.specular_ray, world, depth + 1, 
-                                 backgroundhigh,backgroundlow, tonemap, rng));
+        r1 = r2;
+        if(!diffuse_bounce) {
+          diffuse_bounce = true;
+          r2 = ray(offset_p, p.generate(sampler), r2.pri_stack, r2.time()); //scatters a ray from hit point to stratified direction
+        } else {
+          r2 = ray(offset_p, p.generate(rng), r2.pri_stack, r2.time()); //scatters a ray from hit point to direction
+        }
+        
+        pdf_val = p.value(r2.direction(), rng); //generates a pdf value based the intersection point and the mixture pdf
+        throughput *= hrec.mat_ptr->f(r1, hrec, r2) / pdf_val;
+      } else {
+        return(final_color);
       }
-      cosine_pdf p(hrec.normal);
-      ray scattered = ray(hrec.p, p.generate(rng), r.pri_stack, r.time());
-      pdf_val = p.value(scattered.direction(), rng);
-      return(emitted + srec.attenuation * 
-             hrec.mat_ptr->scattering_pdf(r, hrec, scattered) *  
-             color_amb_uniform(scattered, world, depth + 1, 
-                               backgroundhigh,backgroundlow, tonemap, rng) / pdf_val);
     } else {
-      vec3 unit_direction = unit_vector(r.direction());
-      float t = 0.5 * (unit_direction.y() + 1.0);
-      return(emitted+(1.0 - t) * backgroundlow + t * backgroundhigh);
+      return(final_color);
     }
-  } else {
-    vec3 unit_direction = unit_vector(r.direction());
-    float t = 0.5 * (unit_direction.y() + 1.0);
-    return (1.0 - t) * backgroundlow + t * backgroundhigh;
   }
+#ifdef DEBUG
+  myfile.close();
+#endif
+  return(final_color);
 }
 
 #ifdef DEBUGBVH
 float debug_bvh(const ray& r, hitable *world, random_gen rng) {
   hit_record hrec;
   hrec.bvh_nodes = 0;
-  world->hit(r, 0.000.01, FLT_MAX, hrec, rng);
+  world->hit(r, 0.001, FLT_MAX, hrec, rng);
   return(hrec.bvh_nodes);
 }
 #endif
@@ -252,15 +210,16 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
                        bool progress_bar, int numbercores,
                        bool hasbackground, CharacterVector& background, List& scale_list,
                        NumericVector sigmavec,
-                       float rotate_env, bool verbose, int debug_channel,
+                       float rotate_env, float intensity_env, bool verbose, int debug_channel,
                        IntegerVector& shared_id_mat, LogicalVector& is_shared_mat,
-                       float min_variance, int min_adaptive_size) {
+                       float min_variance, int min_adaptive_size, List glossyinfo,
+                       List image_repeat) {
   auto startfirst = std::chrono::high_resolution_clock::now();
   //Unpack Camera Info
   int nx = as<int>(camera_info["nx"]);
   int ny = as<int>(camera_info["ny"]);
   int ns = as<int>(camera_info["ns"]);
-  Float fov = as<int>(camera_info["fov"]);
+  Float fov = as<Float>(camera_info["fov"]);
   NumericVector lookfromvec = as<NumericVector>(camera_info["lookfrom"]);
   NumericVector lookatvec = as<NumericVector>(camera_info["lookat"]);
   Float aperture = as<Float>(camera_info["aperture"]);
@@ -268,8 +227,11 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
   Float shutteropen = as<Float>(camera_info["shutteropen"]);
   Float shutterclose = as<Float>(camera_info["shutterclose"]);
   Float focus_distance = as<Float>(camera_info["focal_distance"]);
-  int toneval = as<Float>(camera_info["toneval"]);
   NumericVector ortho_dimensions = as<NumericVector>(camera_info["ortho_dimensions"]);
+  size_t max_depth = as<size_t>(camera_info["max_depth"]);
+  size_t roulette_active = as<size_t>(camera_info["roulette_active_depth"]);
+  int sample_method = as<int>(camera_info["sample_method"]);
+  NumericVector stratified_dim = as<NumericVector>(camera_info["stratified_dim"]);
   
   //Initialize output matrices
   NumericMatrix routput(nx,ny);
@@ -281,15 +243,20 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
   vec3 backgroundhigh(bghigh[0],bghigh[1],bghigh[2]);
   vec3 backgroundlow(bglow[0],bglow[1],bglow[2]);
   float dist_to_focus = focus_distance;
-  bool tonemap = toneval == 1 ? false : true;
   CharacterVector alpha_files = as<CharacterVector>(alphalist["alpha_temp_file_names"]);
   LogicalVector has_alpha = as<LogicalVector>(alphalist["alpha_tex_bool"]);
+  
+  CharacterVector bump_files = as<CharacterVector>(alphalist["bump_temp_file_names"]);
+  LogicalVector has_bump = as<LogicalVector>(alphalist["bump_tex_bool"]);
+  NumericVector bump_intensity = as<NumericVector>(alphalist["bump_intensity"]);
+  
+  
   RcppThread::ThreadPool pool(numbercores);
   GetRNGstate();
   random_gen rng(unif_rand() * std::pow(2,32));
   camera cam(lookfrom, lookat, vec3(camera_up(0),camera_up(1),camera_up(2)), fov, float(nx)/float(ny), 
              aperture, dist_to_focus,
-             shutteropen, shutterclose, rng);
+             shutteropen, shutterclose);
   ortho_camera ocam(lookfrom, lookat, vec3(camera_up(0),camera_up(1),camera_up(2)),
                     ortho_dimensions(0), ortho_dimensions(1),
                     shutteropen, shutterclose, rng);
@@ -301,8 +268,12 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
   
   std::vector<Float* > textures;
   std::vector<int* > nx_ny_nn;
+  
   std::vector<Float* > alpha_textures;
   std::vector<int* > nx_ny_nn_alpha;
+  
+  std::vector<Float* > bump_textures;
+  std::vector<int* > nx_ny_nn_bump;
   //Shared material vector
   std::vector<material* >* shared_materials = new std::vector<material* >;
   
@@ -331,6 +302,18 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
       alpha_textures.push_back(nullptr);
       nx_ny_nn_alpha.push_back(nullptr);
     }
+    if(has_bump(i)) {
+      int nxb, nyb, nnb;
+      Float* tex_data_bump = stbi_loadf(bump_files(i), &nxb, &nyb, &nnb, 4);
+      bump_textures.push_back(tex_data_bump);
+      nx_ny_nn_bump.push_back(new int[3]);
+      nx_ny_nn_bump[i][0] = nxb;
+      nx_ny_nn_bump[i][1] = nyb;
+      nx_ny_nn_bump[i][2] = nnb;
+    } else {
+      bump_textures.push_back(nullptr);
+      nx_ny_nn_bump.push_back(nullptr);
+    }
   }
   
   
@@ -342,22 +325,23 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
                                 noise, isnoise, noisephase, noiseintensity, noisecolorlist,
                                 angle, 
                                 isimage, has_alpha, alpha_textures, nx_ny_nn_alpha,
-                                textures, nx_ny_nn,
+                                textures, nx_ny_nn, has_bump, bump_textures, nx_ny_nn_bump,
+                                bump_intensity,
                                 lightintensity, isflipped,
                                 isvolume, voldensity, order_rotation_list, 
                                 isgrouped, group_pivot, group_translate,
                                 group_angle, group_order_rotation, group_scale,
                                 tri_normal_bools, is_tri_color, tri_color_vert, 
                                 fileinfo, filebasedir, 
-                                scale_list, sigmavec, 
-                                shared_id_mat, is_shared_mat, shared_materials, rng);
+                                scale_list, sigmavec, glossyinfo,
+                                shared_id_mat, is_shared_mat, shared_materials,
+                                image_repeat, rng);
   auto finish = std::chrono::high_resolution_clock::now();
   if(verbose) {
     std::chrono::duration<double> elapsed = finish - start;
     Rcpp::Rcout << elapsed.count() << " seconds" << "\n";
   }
   
-
   //Calculate world bounds
   aabb bounding_box_world;
   worldbvh->bounding_box(0,0,bounding_box_world);
@@ -381,38 +365,53 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
   
   if(hasbackground) {
     background_texture_data = stbi_loadf(background[0], &nx1, &ny1, &nn1, 0);
-    background_texture = new image_texture(background_texture_data, nx1, ny1, nn1);
+    background_texture = new image_texture(background_texture_data, nx1, ny1, nn1, 1, 1, intensity_env);
     background_material = new diffuse_light(background_texture);
     background_sphere = new InfiniteAreaLight(nx1, ny1, world_radius*2, world_center,
                                               background_texture, background_material);
     if(rotate_env != 0) {
       background_sphere = new rotate_y(background_sphere, rotate_env);
     }
+  } else if(ambient_light) {
+    //Check if both high and low are black, and set to FLT_MIN
+    if(backgroundhigh.length() == 0 && backgroundlow.length() == 0) {
+      backgroundhigh = vec3(FLT_MIN,FLT_MIN,FLT_MIN);
+      backgroundlow = vec3(FLT_MIN,FLT_MIN,FLT_MIN);
+    }
+    background_texture = new gradient_texture(backgroundlow, backgroundhigh, false, false);
+    background_material = new diffuse_light(background_texture);
+    background_sphere = new InfiniteAreaLight(100, 100, world_radius*2, world_center,
+                                              background_texture, background_material);
   } else {
-    background_texture = new constant_texture(vec3(0,0,0));
-    background_material = new lambertian(background_texture);
-    background_sphere = new sphere(world_center, world_radius, background_material, nullptr);
+    //Minimum intensity FLT_MIN so the CDF isn't NAN
+    background_texture = new constant_texture(vec3(FLT_MIN,FLT_MIN,FLT_MIN));
+    background_material = new diffuse_light(background_texture);
+    background_sphere = new InfiniteAreaLight(100, 100, world_radius*2, world_center,
+                                              background_texture, background_material);
   }
   finish = std::chrono::high_resolution_clock::now();
   if(verbose && hasbackground) {
     std::chrono::duration<double> elapsed = finish - start;
     Rcpp::Rcout << elapsed.count() << " seconds" << "\n";
   }
-  hitable *world_full[2];
-  world_full[0] = worldbvh;
-  world_full[1] = background_sphere;
-  int nval = hasbackground ? 2 : 1;
-  hitable_list world(world_full, nval);
   int numbertosample = 0;
   for(int i = 0; i < implicit_sample.size(); i++) {
     if(implicit_sample(i)) {
       numbertosample++;
     }
   }
-  if(hasbackground) {
-    numbertosample++;
-  }
+  hitable *world_full[2];
+  world_full[0] = worldbvh;
+  world_full[1] = background_sphere;
+  int nval = numbertosample == 0 || hasbackground || ambient_light ? 2 : 1;
   
+  hitable_list world(world_full, nval);
+  bool impl_only_bg = false;
+  if(numbertosample == 0 || hasbackground) {
+    numbertosample++; 
+    impl_only_bg = true;
+  }
+
   std::vector<hitable* > implicit_sample_vector(numbertosample);
   int counter = 0;
   if(verbose) {
@@ -436,24 +435,23 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
     std::chrono::duration<double> elapsed = finish - start;
     Rcpp::Rcout << elapsed.count() << " seconds" << "\n";
   }
-  if(hasbackground) {
+  if(impl_only_bg || hasbackground) {
     implicit_sample_vector[counter] = background_sphere;
   }
-  
 
   hitable_list hlist;
-  if(!implicit_sample_vector.empty()) {
-    hlist = hitable_list(&implicit_sample_vector[0], numbertosample);
-  } else {
-    numbertosample = 0;
-  }
+  hlist = hitable_list(&implicit_sample_vector[0], numbertosample);
+
   if(verbose && !progress_bar) {
     Rcpp::Rcout << "Starting Raytracing:\n ";
   }
+  RProgress::RProgress pb_sampler("Generating Samples [:bar] :percent%");
+  pb_sampler.set_width(70);
   RProgress::RProgress pb("Adaptive Raytracing [:bar] :percent%");
   pb.set_width(70);
 
   if(progress_bar) {
+    pb_sampler.set_total(ny);
     if(min_variance != 0.0f) {
       pb.set_total(ns);
     } else {
@@ -469,7 +467,7 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
         Float v = Float(j) / Float(ny);
         ray r;
         if(fov != 0) {
-          r = cam.get_ray(u,v);
+          r = cam.get_ray(u,v, vec3(0,0,0), 0);
         } else {
           r = ocam.get_ray(u,v);
         }
@@ -488,7 +486,7 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
         Float v = Float(j) / Float(ny);
         ray r;
         if(fov != 0) {
-          r = cam.get_ray(u,v);
+          r = cam.get_ray(u,v, vec3(0,0,0), 0);
         } else {
           r = ocam.get_ray(u,v);
         }
@@ -507,7 +505,7 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
         Float v = Float(j) / Float(ny);
         ray r;
         if(fov != 0) {
-          r = cam.get_ray(u,v);
+          r = cam.get_ray(u,v, vec3(0,0,0), 0);
         } else {
           r = ocam.get_ray(u,v);
         }
@@ -528,7 +526,7 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
         Float v = Float(j + rng.unif_rand()) / Float(ny);
         ray r;
         if(fov != 0) {
-          r = cam.get_ray(u,v);
+          r = cam.get_ray(u,v, vec3(0,0,0));
         } else {
           r = ocam.get_ray(u,v);
         }
@@ -554,48 +552,65 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
       for(int i = 0; i < ny; i++) {
         seeds[i] = unif_rand() * std::pow(2,32);
       }
+      if(verbose) {
+        start = std::chrono::high_resolution_clock::now();
+        if(sample_method == 0) {
+          Rcpp::Rcout << "Allocating random sampler: ";
+        } else {
+          Rcpp::Rcout << "Allocating stratified (" << 
+            stratified_dim(0)<< "x" << stratified_dim(1) << ") sampler: ";
+        }
+      }
+      std::vector<random_gen > rngs;
+      std::vector<Sampler* > samplers;
+      for(int i = 0; i < nx * ny; i++) {
+        random_gen rng_single(unif_rand() * std::pow(2,32));
+        rngs.push_back(rng_single);
+        Sampler* strat;
+        if(sample_method == 0) {
+          strat = new RandomSampler(rng_single);
+        } else {
+          strat = new StratifiedSampler(stratified_dim(0), stratified_dim(1),
+                                        true, 5, rng_single);
+        }
+        strat->StartPixel(vec2(0,0));
+        strat->SetSampleNumber(0);
+        samplers.push_back(strat);
+      }
       RcppThread::ThreadPool pool(numbercores);
       auto worker = [&routput, &goutput, &boutput,
-                     ambient_light, nx, ny, ns, seeds, fov,
-                     &cam, &ocam, backgroundhigh, backgroundlow, &world, &hlist,
-                     numbertosample, clampval, tonemap, progress_bar, 
-                     numbercores, background_texture] (int j) {
+                     nx, ny, ns, &rngs, &samplers, 
+                     fov,
+                     &cam, &ocam, &world, &hlist,
+                     clampval, max_depth, roulette_active, progress_bar, 
+                     numbercores] (int j) {
        if(progress_bar && j % numbercores == 0) {
          RcppThread::Rcout << "Progress (" << numbercores << " core): ";
          RcppThread::Rcout << (int)((1-(double)j/double(ny)) * 100) << "%\r";
        }
-       random_gen rng(seeds[j]);
        std::vector<dielectric*> *mat_stack = new std::vector<dielectric*>;
        
        for(int i = 0; i < nx; i++) {
          vec3 col(0,0,0);
+         int index = j + ny * i;
+         
          for(int s = 0; s < ns; s++) {
-           Float u = Float(i + rng.unif_rand()) / Float(nx);
-           Float v = Float(j + rng.unif_rand()) / Float(ny);
+           vec2 u2 = samplers[index]->Get2D();
+           Float u = (Float(i) + u2.x()) / Float(nx);
+           Float v = (Float(j) + u2.y()) / Float(ny);
            ray r;
            if(fov != 0) {
-             r = cam.get_ray(u,v);
+             r = cam.get_ray(u, v, rand_to_unit(samplers[index]->Get2D()), 
+                             samplers[index]->Get1D());
            } else {
              r = ocam.get_ray(u,v);
            }
            r.pri_stack = mat_stack;
            
-           if(numbertosample) {
-             if(ambient_light) {
-               col += clamp(de_nan(color_amb(r, &world, &hlist, 0,
-                                             backgroundhigh, backgroundlow, tonemap, rng)),0,clampval);
-             } else {
-               col += clamp(de_nan(color(r, &world, &hlist, 0, tonemap, rng)),0,clampval);
-             }
-           } else {
-             if(ambient_light) {
-               col += clamp(de_nan(color_amb_uniform(r, &world, 0, 
-                                                     backgroundhigh, backgroundlow, tonemap, rng)),0,clampval);
-             } else {
-               col += clamp(de_nan(color_uniform(r, &world, 0, tonemap, rng, background_texture)),0,clampval);
-             }
-           }
+           col += clamp(de_nan(color(r, &world, &hlist, max_depth, roulette_active, 
+                                     rngs[index], samplers[index])),0,clampval);
            mat_stack->clear();
+           samplers[index]->StartNextSample();
          }
          col /= Float(ns);
          routput(i,j) = col[0];
@@ -608,6 +623,9 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
         pool.push(worker,j);
       }
       pool.join();
+      for(size_t i = 0; i < samplers.size(); i++) {
+        delete samplers.at(i);
+      }
     } else {
       NumericMatrix routput2(nx,ny);
       NumericMatrix goutput2(nx,ny);
@@ -616,68 +634,88 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
                                               min_variance, min_adaptive_size,
                                               routput, goutput, boutput,
                                               routput2, goutput2, boutput2);
-      for(int s = 0; s < ns; s++) {
+      std::vector<random_gen > rngs;
+      std::vector<Sampler* > samplers;
+      start = std::chrono::high_resolution_clock::now();
+      
+      for(int i = 0; i < nx * ny; i++) {
+        if(progress_bar && (i % ny == 0)) {
+          pb_sampler.tick();
+        }
+        random_gen rng_single(unif_rand() * std::pow(2,32));
+        rngs.push_back(rng_single);
+        Sampler* strat;
+        if(sample_method == 0) {
+          strat = new RandomSampler(rng_single);
+        } else {
+          strat = new StratifiedSampler(stratified_dim(0), stratified_dim(1),
+                                        true, 5, rng_single);
+        }
+        strat->StartPixel(vec2(0,0));
+        strat->SetSampleNumber(0);
+        samplers.push_back(strat);
+      }
+
+      if(verbose) {
+        finish = std::chrono::high_resolution_clock::now();
+        if(sample_method == 0) {
+          Rcpp::Rcout << "Allocating random sampler: ";
+        } else {
+          Rcpp::Rcout << "Allocating stratified (" << 
+            stratified_dim(0)<< "x" << stratified_dim(1) << ") sampler: ";
+        }
+        std::chrono::duration<double> elapsed = finish - start;
+        Rcpp::Rcout << elapsed.count() << " seconds" << "\n";
+      }
+      for(size_t s = 0; s < static_cast<size_t>(ns); s++) {
         Rcpp::checkUserInterrupt();
         if(progress_bar) {
           pb.tick();
         }
-        std::vector<random_gen > rngs;
-        for(int i = 0; i < nx * ny; i++) {
-          random_gen rng_single(unif_rand() * std::pow(2,32));
-          rngs.push_back(rng_single);
-        }
         RcppThread::ThreadPool pool(numbercores);
-        auto worker = [&adaptive_pixel_sampler, 
-                       ambient_light, nx, ny, s, 
-                       &rngs, fov, 
-                       &cam, &ocam, backgroundhigh, backgroundlow, &world, &hlist,
-                       numbertosample, clampval, tonemap, 
-                       background_texture] (int k) {
+        auto worker = [&adaptive_pixel_sampler,
+                       nx, ny, s,
+                       &rngs, fov, &samplers,
+                       &cam, &ocam, &world, &hlist,
+                       clampval, max_depth, roulette_active] (int k) {
            int nx_begin = adaptive_pixel_sampler.pixel_chunks[k].startx;
            int ny_begin = adaptive_pixel_sampler.pixel_chunks[k].starty;
            int nx_end = adaptive_pixel_sampler.pixel_chunks[k].endx;
            int ny_end = adaptive_pixel_sampler.pixel_chunks[k].endy;
-           random_gen rng = rngs[k];
+           
            std::vector<dielectric*> *mat_stack = new std::vector<dielectric*>;
            for(int i = nx_begin; i < nx_end; i++) {
              for(int j = ny_begin; j < ny_end; j++) {
-               vec3 col(0,0,0);
-               Float u = Float(i + rng.unif_rand()) / Float(nx);
-               Float v = Float(j + rng.unif_rand()) / Float(ny);
-               ray r;
+               int index = j + ny * i;
+               vec2 u2 = samplers[index]->Get2D();
+               Float u = (Float(i) + u2.x()) / Float(nx);
+               Float v = (Float(j) + u2.y()) / Float(ny);
+               ray r; 
                if(fov != 0) {
-                 r = cam.get_ray(u,v);
+                 r = cam.get_ray(u, v, rand_to_unit(samplers[index]->Get2D()), 
+                                 samplers[index]->Get1D());
+                 
                } else {
                  r = ocam.get_ray(u,v);
                }
                r.pri_stack = mat_stack;
-               if(numbertosample) {
-                 if(ambient_light) {
-                   col = clamp(de_nan(color_amb(r, &world, &hlist, 0,
-                                                 backgroundhigh, backgroundlow, tonemap, rng)),0,clampval);
-                 } else {
-                   col = clamp(de_nan(color(r, &world, &hlist, 0, tonemap, rng)),0,clampval);
-                 }
-               } else {
-                 if(ambient_light) {
-                   col = clamp(de_nan(color_amb_uniform(r, &world, 0, 
-                                                         backgroundhigh, backgroundlow, tonemap, rng)),0,clampval);
-                 } else {
-                   col = clamp(de_nan(color_uniform(r, &world, 0, tonemap, rng, background_texture)),0,clampval);
-                 }
-               }
+
+               vec3 col = clamp(de_nan(color(r, &world, &hlist, max_depth, 
+                                        roulette_active, rngs[index], samplers[index])),
+                           0, clampval);
                mat_stack->clear();
                adaptive_pixel_sampler.add_color_main(i, j, col);
                if(s % 2 == 0) {
                  adaptive_pixel_sampler.add_color_sec(i, j, col);
                }
+               samplers[index]->StartNextSample();
              }
            }
            if(s % 2 == 1 && s > 1) {
              adaptive_pixel_sampler.test_for_convergence(k, s, nx_end, nx_begin, ny_end, ny_begin);
            }
            delete mat_stack;
-         };
+        };
         for(size_t j = 0; j < adaptive_pixel_sampler.size(); j++) {
           pool.push(worker, j);
         }
@@ -688,6 +726,9 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
         adaptive_pixel_sampler.max_s++;
       }
       adaptive_pixel_sampler.write_final_pixels();
+      for(size_t i = 0; i < samplers.size(); i++) {
+        delete samplers.at(i);
+      }
     }
   }
 
@@ -708,6 +749,10 @@ List render_scene_rcpp(List camera_info, bool ambient_light,
       stbi_image_free(alpha_textures[i]);
       delete nx_ny_nn_alpha[i];
     }
+    if(has_bump(i)) {
+      stbi_image_free(bump_textures[i]);
+      delete nx_ny_nn_bump[i];
+    }
   }
   for(size_t i = 0; i < shared_materials->size(); i++) {
     delete shared_materials->at(i);
